#
# table.pl - show ntop's Pephost hash.
#
# Copyright (C) 2003 Rocco Carbone <rocco@ntop.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING. If not, write to
# the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

#
# the %pephosts hash keeps all information of interest
# for the purpose of reporting statistics access
#
# They include:
#     hwaddress   => unique ethernet address (MAC address)
#     vendor      => unique MAC address vendor name
#     ipaddress   => its IP address in dot notation
#     hostname    => its, locally resolved, hostname
#     domain      => its, locally resolved, DNS domain name
#     nbtname     => its NetBIOS name
#     nbtdomain   => its NetBIOS domain
#     first       => time the host was first seen
#     last        => time the host was last seen

#
# useful variables
#
$__progname__   = "table.pl";
$__version__    = "0.0.1";
$__updated__    = "2003 Mar 10";
$__boottime__   = time();
$__bootdate__   = localtime($__boottime__);

# Hey Netscape, this is for you
my $htmltitle   = "NTOP Statistics -- Traffic Hash";
my $bodyname    = "Traffic Report";
my $title       = "Traffic Summary";
my $description = "Shows the %pephosts hash";

use lib qw(plugins/pep);
require "fmthtml.pl";
require "colorinfo.pl";
require "render.pl";
require "sorting.pl";
require "date.pl";

#
# build various arrays from the %pephosts hash
#
my @vals = values %pephosts;

# sub arrays sorted by various kind of values
my @sorted_by_hwaddress = (sort { $a->{hwaddress} cmp $b->{hwaddress} } @vals)[0..9];
my @sorted_by_vendor = (sort known_vendors @vals)[0..29];

my @sorted_by_ipaddress = (sort ipnumerically @vals)[0..24];

my @unresolved = sort ipnumerically @vals;

my @sorted_by_unresolved = ();
foreach my $s (@unresolved) {
  if ($s->{hostname} =~ /unresolved/) {
    @sorted_by_unresolved = (@sorted_by_unresolved, $s);
  }
}

my @sorted_by_ipless = ();
foreach my $s (@sorted_by_hwaddress) {
  if ($s->{ipaddress} =~ /IP-less/) {
    @sorted_by_ipless = (@sorted_by_ipless, $s);
  }
}

# packets
my @sorted_by_pktsent = (sort { $b->{pktsent} <=> $a->{pktsent} } @vals)[0..9];
my @sorted_by_pktrcvd = (sort { $b->{pktreceived} <=> $a->{pktreceived} } @vals)[0..9];

# bytes
my @sorted_by_bytessent = (sort { $b->{bytessent} <=> $a->{bytessent} } @vals)[0..9];
my @sorted_by_bytesrcvd = (sort { $b->{bytesreceived} <=> $a->{bytesreceived} } @vals)[0..9];

my @sorted_by_lastseen = reverse((sort { $a->{lastseen} <=> $b->{lastseen} } @vals)[0..14]);
my @sorted_by_firstseen = (sort { $a->{firstseen} <=> $b->{firstseen} } @vals)[0..39];
my @sorted_by_firstseen_rev = (reverse (sort { $a->{lastseen} <=> $b->{lastseen} } @vals))[0..24];

my %hosts = (
	     vals     => \@sorted_by_hwaddress,
	     fields   => [
			  "hwaddress", '30%', "left",
			  "hostname",  '30%', "left",
			  "ipaddress", '30%', "right"
			 ],
	     color    => $colorinfo{blue},
	     title    => "Top " . scalar @sorted_by_hwaddress . " Hosts viewed by ntop",
	     subtitle => "Alphabetically sorted by Hardware Address"
	    );

my %unresolved = (
		  vals     => \@sorted_by_unresolved,
		  fields   => [
			       "hwaddress", '50%', "left",
			       "ipaddress", '40%', "left",
			       "pktsent",   '10%', "right",
			       "pktreceived",   '10%', "right",
			       "bytessent",   '10%', "right",
			      ],
		  color    => $colorinfo{red},
		  title    => "Top " . scalar @sorted_by_unresolved . " Hosts viewed by ntop",
		  subtitle => "All the unresolved IP Addresses"
	    );

my %ipless = (
	      vals     => \@sorted_by_ipless,
	      fields   => [
			   "hwaddress", '50%', "left",
			   "hostname",  '50%', "left"
			  ],
	      color    => $colorinfo{red},
	      title    => "Top " . scalar @sorted_by_ipless . " Hosts viewed by ntop",
	      subtitle => "All the IP-less Hardware Addresses"
	     );

my %vendors = (
	       vals     => \@sorted_by_vendor,
	       fields   => [
			    "hwaddress", '30%', "left",
			    "vendor",    '30%', "center",
			    "ipaddress", '30%', "right"
			   ],
	       color    => $colorinfo{blue},
	       title    => "Top " . scalar @sorted_by_vendor . " Hosts viewed by ntop",
	       subtitle => "Alphabetically sorted by Vendor Name"
	      );

my %ips = (
	   vals     => \@sorted_by_ipaddress,
	   fields   => [
#			"hwaddress", '30%', "left",
			"ipaddress", '30%', "left",
			"hostname",  '30%', "left",
			"pktsent",   '30%', "right"
		       ],
	   color    => $colorinfo{yellow},
	   title    => "Top " . scalar @sorted_by_ipaddress . " Hosts viewed by ntop",
	   subtitle => "Numerically sorted by the IP Address"
	  );

my %pktsent = (
	       vals     => \@sorted_by_pktsent,
	       fields   => [
			    "hwaddress", '30%', "left",
			    "hostname",  '30%', "left",
			    "pktsent",   '30%', "right"
			   ],
	       color    => $colorinfo{red},
	       title    => "Top " . scalar @sorted_by_pktsent . " Hosts viewed by ntop",
	       subtitle => "Numerically sorted by the # of packet sent"
	     );

my %pktrcvd = (
	       vals     => \@sorted_by_pktrcvd,
	       fields   => [
			    "hwaddress",   '30%', "left",
			    "hostname",    '30%', "left",
			    "pktreceived", '30%', "right"
			   ],
	       color    => $colorinfo{red},
	       title    => "Top " . scalar @sorted_by_pktrcvd . " Hosts viewed by ntop",
	       subtitle => "Numerically sorted by the # of packet received"
	     );

my %bytessent = (
		 vals     => \@sorted_by_bytessent,
		 fields   => [
			      "hwaddress", '30%', "left",
			      "hostname",  '30%', "left",
			      "bytessent", '30%', "right"
			     ],
		 color    => $colorinfo{green},
		 title    => "Top " . scalar @sorted_by_bytessent . " Hosts viewed by ntop",
		 subtitle => "Numerically sorted by the # of bytes sent"
	     );

my %bytesrcvd = (
		 vals     => \@sorted_by_bytesrcvd,
		 fields   => [
			      "hwaddress",     '30%', "left",
			      "hostname",      '30%', "left",
			      "bytesreceived", '30%', "right"
			     ],
		 color    => $colorinfo{green},
		 title    => "Top " . scalar @sorted_by_bytesrcvd . " Hosts viewed by ntop",
		 subtitle => "Numerically sorted by the # of bytes received"
	     );

my %last  = (
	     vals     => \@sorted_by_lastseen,
	     fields   => [
			  "hwaddress", '30%', "left",
			  "hostname",  '30%', "left",
			  "lastseen",  '30%', "center"
			 ],
	     color    => $colorinfo{green},
	     title    => "Top " . scalar @sorted_by_lastseen . " Hosts viewed by ntop",
	     subtitle => "Alphabetically sorted by Last Seen"
	    );

my %first = (
	     vals     => \@sorted_by_firstseen,
	     fields   => [
			  "hwaddress", '30%', "left",
			  "hostname",  '30%', "left",
			  "firstseen", '30%', "center"
			 ],
	     color    => $colorinfo{yellow},
	     title    => "Top " . scalar @sorted_by_firstseen . " Hosts viewed by ntop",
	     subtitle => "Alphabetically sorted by First Seen"
	    );

my %first_rev = (
		 vals     => \@sorted_by_firstseen_rev,
		 fields   => [
			      "hwaddress", '30%', "left",
			      "hostname",  '30%', "left",
			      "firstseen", '30%', "center"
			     ],
		 color    => $colorinfo{red},
		 title    => "Top " . scalar @sorted_by_firstseen_rev . " Hosts viewed by ntop",
		 subtitle => "Alphabetically sorted by First Seen (reverse)"
		);

my @views = (\%unresolved, \%ipless,
	     \%hosts, \%vendors,
	     \%ips,
	     \%pktsent, \%pktrcvd,
	     \%bytessent, \%bytesrcvd,
	     \%last, \%first, \%first_rev);


#
# yet another main here
#
select STDOUT;

#
# print HTML header and title
#
html_header($htmltitle);

#
# start printing HTML body
#
html_start_body($bodyname);

#
# print report's title with its brief description
#
print "<center><font size=6><b>$description</b></font></center>\n";

#
# print the matrix
#
print_matrix(\@views, 2, 400);

#
# finish HTML output
#
html_stop_body();

1;
